/* SMTPŒPɃ[]     */
/* Copyright (c) 2018 TODA, Takashi */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>
#ifdef	WIN32
#include <io.h>
#include <winsock.h>
#include <conio.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#endif
#include <time.h>
#include <errno.h>
#ifdef	MSDOS
#include <dos.h>
#endif

#ifdef	WIN32
#define EWOULDBLOCK	WSAEWOULDBLOCK
#define	perror(routine)		fprintf(stderr, "%s: Error Code %d\n", routine, WSAGetLastError())
#endif

#ifdef	MSDOS
#define	O_NDELAY	FNDELAY
#endif

/* Validate Next Line to use SMTP-AUTH */ 
/* #define	SMTPAUTH */

#ifdef	SMTPAUTH
#define	PSWDOPT	2
#else
#define	PSWDOPT	0
#endif

/* Name of Protocol defined in /etc/services */
#ifdef	SMTPAUTH
#define	PROTOCOL	"submission"
#define	PROTOCOLNUM	587
#else
#define	PROTOCOL	"smtp"
#define	PROTOCOLNUM	25
#endif
/* Validate Next Line to distinguish Protocol by Name */ 
#define	PROTOCOLBYNAME

/* Size of Socket Buffer */
#define SOCKBUFSIZE		4096

/* Size of Line Buffer */
#define	LINEBUFSIZE		4096

/* Size of Address Buffer */
#define	ADDRBUFSIZE		4096

/* Size of Password-Encode Buffer */
#define	PSWDBUFSIZE		1024

/* Prototype Definitions */
static int	ConnectServer(char *);
static void	DisconnectServer(int);
static int	GetReply(void);
static void	SendCommand(void);
static char	*ReceiveCommand(void);
#ifdef	SMTPAUTH
static void	putb64(int), endb64(void);
#endif

/* Global Variables */
static FILE	*fp;
static char	*message, *sockbuf, *addrbuf;
static int	sock;
#ifdef	SMTPAUTH
static unsigned char	b64_buffer[3] = {0, 0, 0};
static int	b64ptr = 0;
static char	b64_table[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static char	*pswdbuf, *pswdptr;
#endif

/* Command Line Options */
static int	verbose = 0;
static int	my_optind;	/* in stead of 'optind' with getopt() */

int	main(int argc, char *argv[]){
	int	c;
	char	*ptr;
#ifdef	WIN32
	WSADATA	wsaData;
#endif

#ifdef	WIN32
	if( WSAStartup(MAKEWORD(1, 1), &wsaData)){
		fprintf(stderr, "WinSock unavailiable\n");
		exit(126);
	}
#endif

	if( (message = malloc(LINEBUFSIZE)) == NULL ){
		fprintf(stderr, "Cannot Allocate Transmit Line Buffer.\n");
		exit(126);
	}
	if( (sockbuf = malloc(SOCKBUFSIZE)) == NULL ){
		fprintf(stderr, "Cannot Allocate Transmit Socket Buffer.\n");
		exit(126);
	}
	if( (addrbuf = malloc(ADDRBUFSIZE)) == NULL ){
		fprintf(stderr, "Cannot Allocate Address Analyze Buffer.\n");
		exit(126);
	}
#ifdef	SMTPAUTH
	if( (pswdbuf = malloc(PSWDBUFSIZE)) == NULL ){
		fprintf(stderr, "Cannot Allocate Password Encode Buffer.\n");
		exit(126);
	}
	pswdptr = pswdbuf;
#endif

	/* some of the author's targets do not support getopt()  */
	my_optind = 1;
	if( argc <= 1 ){
		my_optind = -(6 + PSWDOPT);
	}
	else if( argv[1][0] == '-' ){
		if( argv[1][2] ){
			my_optind = -(6 + PSWDOPT);
		}
		else if( argv[1][1] == '-' ){
			my_optind++;
		}
		else if( argv[1][1] == 'v' ){
			verbose++;
			my_optind++;
		}
		else{
			my_optind = -(6 + PSWDOPT);
		}
	}
	if( (argc > my_optind + 5 + PSWDOPT) || (argc <= my_optind + 3 + PSWDOPT) ){
#ifdef	SMTPAUTH
		fprintf(stderr, "Usage: %s [ -v ] hostname account password myhostname fromaddr sendaddr [mailfile]\n", argv[0]);
#else
		fprintf(stderr, "Usage: %s [ -v ] hostname myhostname fromaddr sendaddr [mailfile]\n", argv[0]);
#endif
		exit(4);
	}

	if( argc != my_optind + 5 + PSWDOPT){
		fp = stdin;
	}
	else if( (fp = fopen(argv[my_optind + 4 + PSWDOPT], "r")) == NULL ){
		fprintf(stderr, "cannot open Mail Data File '%s'.\n", argv[my_optind + 4 + PSWDOPT]);
		exit(126);
	}

#ifdef	SMTPAUTH
	for( ptr = argv[my_optind + 1] ; *ptr ; ptr++ ){
		putb64(*ptr);
		if( pswdptr - pswdbuf > PSWDBUFSIZE - 5 ){
			fprintf(stderr, "Too Long Encoded Password.\n");
			exit(126);
		}
	}
	putb64(0);
	for( ptr = argv[my_optind + 1] ; *ptr ; ptr++ ){
		putb64(*ptr);
		if( pswdptr - pswdbuf > PSWDBUFSIZE - 5 ){
			fprintf(stderr, "Too Long Encoded Password.\n");
			exit(126);
		}
	}
	putb64(0);
	for( ptr = argv[my_optind + 2] ; *ptr ; ptr++ ){
		putb64(*ptr);
		if( pswdptr - pswdbuf > PSWDBUFSIZE - 5 ){
			fprintf(stderr, "Too Long Encoded Password.\n");
			exit(126);
		}
	}
	endb64();
#endif

	if( (sock = ConnectServer(argv[my_optind])) < 0 ){
		fprintf(stderr, "cannot connect '%s'\n", argv[my_optind]);
		exit(126);
	}
	if( GetReply() >= 300 ){
		fprintf(stderr, "connect: %s\n", message);
		DisconnectServer(0);
		exit(126);
	}

	strcpy(message, "HELO ");
	strcat(message, argv[my_optind + 1 + PSWDOPT]);	/* FQDN of the sender */
	SendCommand();
	if( GetReply() >= 300 ){
		fprintf(stderr, "HELO: %s\n", message);
		DisconnectServer(0);
		exit(62);
	}

#ifdef	SMTPAUTH
	strcpy(message, "AUTH PLAIN ");
	strcat(message, pswdbuf);
	SendCommand();
	if( GetReply() >= 300 ){
		fprintf(stderr, "AUTH: %s\n", message);
		DisconnectServer(0);
		exit(62);
	}
#endif

	strcpy(message, "MAIL FROM: ");
	strcat(message, argv[my_optind + 2 + PSWDOPT]);	/* Envelope "From Adreess" */
	SendCommand();
	if( GetReply() >= 300 ){
		fprintf(stderr, "MAIL FROM: %s\n", message);
		DisconnectServer(0);
		exit(62);
	}

	strcpy(message, "RCPT TO: <");
	strcat(message, argv[my_optind + 3 + PSWDOPT]);	/* Destinate Adreess */
	strcat(message, ">");	/* Destinate Adreess */
	SendCommand();
	if( GetReply() > 300 ){
		fprintf(stderr, "RCPT TO: %s\n", message);
		DisconnectServer(0);
		exit(62);
	}

	strcpy(message, "DATA");
	SendCommand();
	if( GetReply() != 354 ){
		fprintf(stderr, "DATA: %s\n", message);
		DisconnectServer(0);
		exit(62);
	}

	c = 0;
	while( c != EOF ){
		ptr = message;
		c = getc(fp);
		if( c == '.' ){
			*ptr++ = '.';
		}
		while( (c != '\n') && (c != EOF) && ((ptr - message) < LINEBUFSIZE - 4) ){
			*ptr++ = c;
			c = getc(fp);
		}
		if( (c != EOF) || (ptr != message) ){
			*ptr++ = 0;
			SendCommand();
		}
	}

	strcpy(message, ".");
	SendCommand();
	if( GetReply() >= 300 ){
		fprintf(stderr, "End of DATA: %s\n", message);
		DisconnectServer(0);
		exit(62);
	}
	fputs(message, stderr);
	putc('\n', stderr);

	DisconnectServer(1);
	return	0;
}

static int	ConnectServer(char *hostname){
	struct sockaddr_in	server;
#ifdef	PROTOCOLBYNAME
	struct servent	*sp;
#endif
	struct hostent	*hp;
	int	sock;

	memset((char *)&server, 0, sizeof(server));
	server.sin_addr.s_addr = inet_addr(hostname);
	if( server.sin_addr.s_addr != ~0UL ){
		server.sin_family = AF_INET;
	}
	else{
		hp = gethostbyname(hostname);
		if( hp == NULL ){
			fprintf(stderr, "unknown host: %s\n", hostname);
			return	-1;
		}
		server.sin_family = hp->h_addrtype;
		memcpy((char *)&server.sin_addr, hp->h_addr, (size_t)hp->h_length);
	}
#ifdef	PROTOCOLBYNAME
	sp = getservbyname(PROTOCOL, "tcp");
	if( sp == NULL ){
		fprintf(stderr, "unknown protocol: %s\n", PROTOCOL);
		return	-1;
	}
	server.sin_port = sp->s_port;
#else
	server.sin_port = htons(PROTOCOLNUM);
#endif
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if( sock < 0 ){
		perror("socket");
		return	-1;
	}
	if( connect(sock, (struct sockaddr *)&server, sizeof(server)) < 0 ){
		perror("connect");
		return	-1;
	}
#ifdef	MSDOS
	if( fcntl(sock, F_SETFL, O_NDELAY) ){
		perror("fcntl");
		return	-1;
	}
#endif
	return	sock;
}

static void DisconnectServer(int waitreply){
	strcpy(message, "QUIT");
	SendCommand();
	if( waitreply ){
		if( GetReply() >= 300 ){
			fprintf(stderr, "QUIT: %s\n", message);
			DisconnectServer(0);
		}
	}
#ifdef	WIN32
	closesocket(sock);
#else
	close(sock);
#endif
}

static int	GetReply(void){
	strcpy(message, "000-");
	while( message[3] == '-' ){
		if( ReceiveCommand() == NULL ){
			return	0x7fff;
		}
		if( verbose ){
			fputs(message, stderr);
			putc('\n', stderr);
		}
	}
	return	(((message[0] - '0') & 0xff) * 100) + (((message[1] - '0') & 0xff) * 10) + ((message[2] - '0') & 0xff);
}

static void	SendCommand(void){
	strcat(message, "\r\n");
	send(sock, message, strlen(message), 0);
	if( verbose ){
		message[strlen(message) - 2] = '\n';
		message[strlen(message) - 1] = 0;
		fputs(message, stderr);
	}
}

static char	*ReceiveCommand(void){
	static	int	recvret = 0, reststart = 0;
	int	linelength = 0;

	for(;;){
		while( recvret > reststart ){
			message[linelength] = sockbuf[reststart];
			reststart++;
			if( (message[linelength] == '\n') &&
			    (linelength > 0) &&
			    (message[linelength - 1] == '\r') ){
				message[linelength - 1] = 0;
				return	message;
			}
			if( message[linelength] == 0 ){
				return	message;
			}
			if( ++linelength >= LINEBUFSIZE - 1 ){
				message[linelength] = 0;
				return	message;
			}
		}
		reststart = 0;
		if( (recvret = recv(sock, sockbuf, SOCKBUFSIZE, 0)) == 0 ){
			fprintf(stderr, "recv: Unexpected Disconnect\n");
			return	NULL;
		}
		else if( recvret < 0 ){
#ifdef	WIN32
			if( WSAGetLastError() != EWOULDBLOCK ){
#else
			if( errno != EWOULDBLOCK ){
#endif
				perror("recv");
				return	NULL;
			}
#ifdef	MSDOS
			else if( (bdos(6, 0xff, 0) & 0xff) == 0x03 ){
				fprintf(stderr, "^C\n");
				return	NULL;
			}
#endif
		}
	}
}

#ifdef	SMTPAUTH
static	void	putb64(int c)
{
	b64_buffer[b64ptr++] = c;
	if( b64ptr >= 3 ){
		*pswdptr++ = b64_table[(b64_buffer[0] >> 2)  & 0x3f];
		*pswdptr++ = b64_table[((b64_buffer[0] << 4) | (b64_buffer[1] >> 4)) & 0x3f];
		*pswdptr++ = b64_table[((b64_buffer[1] << 2) | (b64_buffer[2] >> 6)) & 0x3f];
		*pswdptr++ = b64_table[b64_buffer[2] & 0x3f];
		for( b64ptr = 0 ; b64ptr < 3 ; b64ptr++ ){
			b64_buffer[b64ptr] = 0;
		}
		b64ptr = 0;
	}
}

static	void	endb64()
{
	if( b64ptr > 0 ){
		*pswdptr++ = b64_table[(b64_buffer[0] >> 2)  & 0x3f];
		*pswdptr++ = b64_table[((b64_buffer[0] << 4) | (b64_buffer[1] >> 4)) & 0x3f];
		*pswdptr++ = (b64ptr == 1) ? '=' : b64_table[((b64_buffer[1] << 2) | (b64_buffer[2] >> 6)) & 0x3f];
		*pswdptr++ = '=';
	}
		*pswdptr++ = 0;
}
#endif
